"""Optimization procedure for iterative methods (e.g. gradient descent)."""

from logging import Logger, root, INFO
from typing import Dict, Any, List, Tuple, Callable, Optional

from tqdm.auto import tqdm  # type: ignore

import lab

from scnn.private.methods.optimization_procedures.optimization_procedure import (
    OptimizationProcedure,
    ITER_LOG_FREQ,
    METRIC_FREQ,
)

from scnn.private.methods.optimizers import Optimizer, ProximalOptimizer
from scnn.private.models.model import Model
from scnn.private.metrics import (
    update_metrics,
    init_metrics,
    format_recent_metrics,
    merge_metric_lists,
)
from scnn.private.methods.termination_criteria import TerminationCriterion


class IterativeOptimizationProcedure(OptimizationProcedure):

    """An iterative optimization procedure."""

    optimizer: Optimizer

    def __init__(
        self,
        optimizer: Optimizer,
        max_iters: int,
        term_criterion: TerminationCriterion,
        name: str = "",
        divergence_check: bool = True,
        batch_size: Optional[int] = None,
        log_freq: int = ITER_LOG_FREQ,
        metric_freq: int = METRIC_FREQ,
        pre_process: Optional[Callable] = None,
        post_process: Optional[Callable] = None,
    ):
        """
        :param optimizer: an optimizer instance implementing the 'step' method.
        :param max_iters: the maximum number of iterations to run the optimizer.
        :param term_criterion: the criterion to use when checking for early termination.
        :param name: an optional name to display when printing the progress bar.
        :param divergence_check: an whether or not the optimization procedure should
            check for divergence behavior and terminate early.
        :param batch_size: the batch size to use when computing the objective.
            Defaults to `None' which indicates full-batch.
        :param log_freq: the frequency at which to information during optimization.
        :param metric_freq: the frequency at which to collect metrics, like objective,
            gradient-norm, etc, during optimization.
        :param pre_process: (optional) a function to call on the model after the it is
            initialized but *before* optimization starts.
        :param pre_process: (optional) a function to call on the model *after* optimization
            is complete.
        """
        self.optimizer = optimizer
        self.max_iters = max_iters
        self.term_criterion = term_criterion

        self.name = name
        self.divergence_check = divergence_check
        self.batch_size = batch_size

        self.log_freq = log_freq
        self.metric_freq = metric_freq

        self.pre_process = pre_process
        self.post_process = post_process

    def reset(self):
        """Reset the optimization procedure (including any attached optimizers)
        to their original state."""
        self.optimizer.reset()

    def _pre_optimization(
        self,
        logger: Logger,
        model: Model,
        initializer: Callable[[Model], Model],
        train_data: Tuple[lab.Tensor, lab.Tensor],
        test_data: Tuple[lab.Tensor, lab.Tensor],
        metrics: Tuple[List[str], List[str], List[str]],
        final_metrics: Optional[Tuple[List[str], List[str], List[str]]],
        step_size: float,
    ) -> Tuple[Model, Dict[str, Any], Dict[str, Any], float, lab.Tensor]:

        exit_status: Dict[str, Any] = {}
        metrics_log: Dict[str, Any] = init_metrics(
            {}, merge_metric_lists(metrics, final_metrics)
        )
        X, y = train_data

        # initialize model
        start_time = self._get_start_time(None)
        model = initializer(model)

        # optional pre-processing.
        if self.pre_process is not None:
            model = self.pre_process(model, X, y)

        # compute initial objective and gradient.
        objective = model.objective(X, y, batch_size=self.batch_size)
        grad = model.grad(
            X,
            y,
            batch_size=self.batch_size,
            step_size=step_size,
        )

        # initial metrics
        metrics_log, _ = self._record_time(metrics_log, start_time)

        metrics_log = update_metrics(
            metrics_log,
            model,
            {"step_size": step_size},
            train_data,
            test_data,
            metrics,
            objective,
            grad,
            batch_size=self.batch_size,
        )
        logger.info(
            "Pre-Optimization Metrics: "
            + format_recent_metrics(metrics_log, metrics)
        )

        return (
            model,
            metrics_log,
            exit_status,
            objective,
            grad,
        )

    def _post_optimization(
        self,
        logger: Logger,
        model: Model,
        train_data: Tuple[lab.Tensor, lab.Tensor],
        test_data: Tuple[lab.Tensor, lab.Tensor],
        metrics: Tuple[List[str], List[str], List[str]],
        final_metrics: Optional[Tuple[List[str], List[str], List[str]]],
        exit_status: Dict[str, Any],
        metrics_log: Dict[str, Any],
        start_time: float,
        step_size: float,
    ) -> Tuple[Model, Dict[str, Any], Dict[str, Any]]:

        X, y = train_data

        start_time = self._get_start_time(start_time)

        # optional post-processing.
        if self.post_process is not None:
            model = self.post_process(model, X, y)

        metrics_log, start_time = self._record_time(metrics_log, start_time)

        # compute post-training metrics.
        objective = model.objective(X, y, batch_size=self.batch_size)
        grad = model.grad(X, y, batch_size=self.batch_size)

        metrics_log = update_metrics(
            metrics_log,
            model,
            {"step_size": step_size},
            train_data,
            test_data,
            merge_metric_lists(metrics, final_metrics),
            objective,
            grad,
            batch_size=self.batch_size,
        )

        logger.info(
            "Post-Optimization Metrics: "
            + format_recent_metrics(metrics_log, metrics)
        )

        return model, exit_status, metrics_log

    def __call__(
        self,
        logger: Logger,
        model: Model,
        initializer: Callable[[Model], Model],
        train_data: Tuple[lab.Tensor, lab.Tensor],
        test_data: Tuple[lab.Tensor, lab.Tensor],
        metrics: Tuple[List[str], List[str], List[str]],
        final_metrics: Optional[Tuple[List[str], List[str], List[str]]] = None,
        callback: Callable = None,
    ) -> Tuple[Dict[str, Any], Model, Dict[str, Any]]:
        """Optimize a model using an iterative optimize method by running a
        "optimization" loop. The loop runs for 'max_iters' number of
        iterations, collects optional metrics at every iteration, and will
        terminate early if the conditions of 'term_criterion' are met.

        :param logger: a logging.Logger object that can be used to log information during
            optimization.
        :param model: the model to optimize.
        :param initializer: a function that can be called to initialize the model.
        :param train_data: an (X, y) tuple representing the training set.
        :param test_data: an (X_test, y_test) tuple representing the test data.
        :param metrics: a tuple of the form (train_metrics, test_metrics, additional_metrics)
            specifying the metrics to be computed on the training set, test set, and data-independent
            metrics.
        :returns: (exit_status, model, metrics) --- execution information, the optimized model, and metrics from optimization.
        """

        # pre-optimization setup
        (
            model,
            metrics_log,
            exit_status,
            objective,
            grad,
        ) = self._pre_optimization(
            logger,
            model,
            initializer,
            train_data,
            test_data,
            metrics,
            final_metrics,
            self.optimizer.step_size,
        )
        (X, y), f0, start_time = train_data, objective, None

        # optimization loop
        verbose = root.level <= INFO
        for itr in tqdm(
            range(self.max_iters), desc=self.name, disable=(not verbose)
        ):
            start_time = self._get_start_time(start_time)
            if itr % self.log_freq == 0 and verbose:
                tqdm.write(format_recent_metrics(metrics_log, metrics))

            # check termination criteria
            if self.term_criterion(model, X, y, objective, grad):
                exit_status["success"] = True
                logger.info(
                    f"Termination criterion satisfied at iteration {itr}/{self.max_iters}. Exiting optimization loop."
                )
                break

            # check for divergence
            if self.divergence_check and objective > 1000 * f0:
                exit_status["success"] = False
                logger.warning(
                    f"Method diverged at {itr}/{self.max_iters}. Exiting optimization loop."
                )
                break

            model, objective, sp_exit_state = self.optimizer.step(
                model, X, y, objective, grad, batch_size=self.batch_size
            )

            if callback is not None:
                model = callback(model, X, y)

            # compute objective and gradient for next iteration.
            if objective is None:
                objective = model.objective(X, y, batch_size=self.batch_size)

            grad = model.grad(
                X,
                y,
                batch_size=self.batch_size,
                step_size=self.optimizer.step_size,
            )

            # calculate time and other metrics.
            if itr % self.metric_freq == 0:
                metrics_log, start_time = self._record_time(
                    metrics_log, start_time
                )
                metrics_log = update_metrics(
                    metrics_log,
                    model,
                    sp_exit_state,
                    train_data,
                    test_data,
                    metrics,
                    objective,
                    grad,
                    batch_size=self.batch_size,
                )

        # post-optimization clean-up.
        model, exit_status, metrics_log = self._post_optimization(
            logger,
            model,
            train_data,
            test_data,
            metrics,
            final_metrics,
            exit_status,
            metrics_log,
            start_time,
            self.optimizer.step_size,
        )

        # collect run-information
        if itr == self.max_iters - 1:
            exit_status["success"] = False
            logger.warning(
                "Max iterations reached before termination criterion was satisfied. Exiting optimization loop."
            )
        exit_status["iterations"] = itr + 1

        return exit_status, model, metrics_log



"""Group L1 Regularizer."""

from typing import Optional

import lab

from scnn.private.loss_functions import group_l1_penalty
from scnn.private.models.regularizers.regularizer import Regularizer


class GroupL1Regularizer(Regularizer):

    """The group-sparsity inducing Group L1-regularizer.

    The group-L1 regularizer, sometimes called the L1-L2 regularizer,
    has the mathematical form,

        ::math.. R(w) = \\lambda \\sum_{i \\in \\mathcal{I} ||w_i||_2,

    where :math:`\\mathcal{I}` is collection of disjoint index sets specifying
    the groups for the regularizer. This class expects the final axis of the
    inputs/weights :math:`w` to be the group axis.

    The group-L1 regularizer induces group-sparsity, meaning entire groups
    :math:`w_i` will be set to zero when :math:`\\lambda` is sufficiently
    large.

    Attributes:
        lam: the regularization strength.

    """

    def __init__(self, lam: float):
        """
        lam: a tuning parameter controlling the regularization strength.
        """

        self.lam = lam

    def penalty(
        self,
        w: lab.Tensor,
        **kwargs,
    ) -> float:
        """Compute the penalty associated with the regularizer.

        Args:
            w: parameter at which to compute the penalty.

        Returns:
            The value of the penalty at w.
        """
        return group_l1_penalty(w, self.lam)

    def grad(
        self,
        w: lab.Tensor,
        base_grad: Optional[lab.Tensor] = None,
        **kwargs,
    ) -> lab.Tensor:
        """Compute the minimum-norm subgradient (aka, the pseudo-gradient).

        Args:
            w: parameter at which to compute the penalty gradient.
            base_grad: the gradient of the un-regularized objective.
                This is required to compute the minimum-norm subgradient.

        Returns:
            minimum-norm subgradient.
        """
        # requires base_grad to compute minimum-norm subgradient
        assert base_grad is not None

        weight_norms = lab.sqrt(lab.sum(w ** 2, axis=-1, keepdims=True))
        grad_norms = lab.sqrt(lab.sum(base_grad ** 2, axis=-1, keepdims=True))

        # TODO: use safe divide
        non_smooth_term = (
            base_grad
            * lab.smin(self.lam / grad_norms, 1)
            * (weight_norms == 0)
        )
        smooth_term = self.lam * lab.safe_divide(w, weight_norms)

        # match input shape
        subgrad = smooth_term - non_smooth_term

        return subgrad



"""Rules for update step-sizes after a line-search."""
from typing import Optional

import lab


class StepSizeUpdater:

    """Base class for methods to update step-sizes between iterations."""

    def __call__(
        self,
        new_step_size: float,
        old_step_size: float,
        s: Optional[lab.Tensor] = None,
        f0: Optional[float] = None,
        f1: Optional[float] = None,
        grad: Optional[lab.Tensor] = None,
    ) -> float:
        """Compute a new step-size given the result of the most recent line-
        search and the step-size from the previous iteration.

        :param new_step_size: the step-size most recently returned by the line-search.
        :param old_step-size: the step-size from the previous iteration.
        :param s: the difference in iterates: w_k - w_{k-1}
        """

        raise NotImplementedError("A step-size update must implement '__call__'!")


class KeepNew(StepSizeUpdater):

    """Simple update rule which returns step-size from the most recent line-
    search."""

    def __call__(
        self,
        new_step_size: float,
        old_step_size: float,
        s: Optional[lab.Tensor] = None,
        f0: Optional[float] = None,
        f1: Optional[float] = None,
        grad: Optional[lab.Tensor] = None,
    ) -> float:
        """Compute a new step-size given the result of the most recent line-
        search and the step-size from the previous iteration.

        :param new_step_size: the step-size most recently returned by the line-search.
        :param old_step-size: the step-size from the previous iteration.
        :param s: the difference in iterates: w_k - w_{k-1}
        """

        return new_step_size


class KeepOld(StepSizeUpdater):

    """Simple update rule which returns the old step-size."""

    def __call__(
        self,
        new_step_size: float,
        old_step_size: float,
        s: Optional[lab.Tensor] = None,
        f0: Optional[float] = None,
        f1: Optional[float] = None,
        grad: Optional[lab.Tensor] = None,
    ) -> float:
        """Compute a new step-size given the result of the most recent line-
        search and the step-size from the previous iteration.

        :param new_step_size: the step-size most recently returned by the line-search.
        :param old_step-size: the step-size from the previous iteration.
        :param s: the difference in iterates: w_k - w_{k-1}
        """

        return old_step_size


class ForwardTrack(StepSizeUpdater):

    """Compute the new step-size by slightly increasing the value returned by
    the line-search."""

    def __init__(self, alpha: float = 1.1):
        """
        :param alpha: the amount to multiplicatively increase the step-size by at each iteration.
        """
        self.alpha = alpha

    def __call__(
        self,
        new_step_size: float,
        old_step_size: float,
        s: Optional[lab.Tensor] = None,
        f0: Optional[float] = None,
        f1: Optional[float] = None,
        grad: Optional[lab.Tensor] = None,
    ) -> float:
        """Compute a new step-size given the result of the most recent line-
        search and the step-size from the previous iteration.

        :param new_step_size: the step-size most recently returned by the line-search.
        :param old_step-size: the step-size from the previous iteration.
        :param s: the difference in iterates: w_k - w_{k-1}
        """

        return new_step_size * self.alpha


class Lassplore(StepSizeUpdater):

    """Step-size update rule proposed in 'Large Scale Spare Logistic
    Regression' by Liu et al.

    [https://dl.acm.org/doi/abs/10.1145/1557019.1557082]
    """

    def __init__(self, alpha: float = 1.2, threshold: float = 5.0):
        """
        :param alpha: the amount to multiplicatively increase the step-size by at each iteration.
        :param threshold: threshold for increasing the step-size after a successful line-search.
        """
        self.alpha = alpha
        self.threshold = threshold

    def __call__(
        self,
        new_step_size: float,
        old_step_size: float,
        s: Optional[lab.Tensor] = None,
        f0: Optional[float] = None,
        f1: Optional[float] = None,
        grad: Optional[lab.Tensor] = None,
    ) -> float:
        """Compute a new step-size given the result of the most recent line-
        search and the step-size from the previous iteration.

        :param new_step_size: the step-size most recently returned by the line-search.
        :param old_step-size: the step-size from the previous iteration.
        :param s: the difference in iterates: w_k - w_{k-1}
        :param grad: the current gradient g_k.
        """
        # required arguments
        assert f0 is not None
        assert f1 is not None
        assert grad is not None
        assert s is not None

        denom = (f1 - (f0 + lab.sum(lab.multiply(grad, s)))) * 2 * new_step_size
        gap = lab.sum(s ** 2)

        if gap > self.threshold * denom:
            return new_step_size * self.alpha
        else:
            return new_step_size



"""Core optimization routines."""

from .line_search import ls
from .gradient_descent import gradient_step, gd_ls
from .proximal_gradient import (
    proximal_gradient_step,
    proximal_gradient_ls,
    fista_step,
    fista_ls,
)
from .augmented_lagrangian import update_multipliers, acc_update_multipliers


__all__ = [
    "ls",
    "gradient_step",
    "gd_ls",
    "proximal_gradient_step",
    "proximal_gradient_ls",
    "fista_step",
    "fista_ls",
    "update_multipliers",
    "acc_update_multipliers",
]



"""The FISTA accelerated proximal gradient descent method."""
from typing import Optional, cast, Dict, Any, Tuple

import lab

from scnn.private.methods.core.proximal_gradient import (
    fista_ls,
)
from scnn.private.models.model import Model
from scnn.private.methods.optimizers.proximal_optimizer import ProximalLSOptimizer

from scnn.private.prox import ProximalOperator
from scnn.private.methods.line_search import StepSizeUpdater, LSCondition, Backtracker

# constants
F_VALS: str = "function_values"
GM: str = "gradient_mapping"


class FISTA(ProximalLSOptimizer):

    """Fast iterative shrinkage-thresholding method."""

    def __init__(
        self,
        init_step_size: float,
        ls_cond: LSCondition,
        backtrack_fn: Backtracker,
        update_step_size: StepSizeUpdater,
        prox: ProximalOperator,
        mu: float = 0.0,
        restart_rule: Optional[str] = GM,
    ):
        """
        :param step_fn: a function to call to execute one step of the iterative
            method. It should have the same signature as 'methods.gradient_descent.gd_ls'.
        :param init_step_size: first step-size to try when running the line-search.
        :param ls_cond: the line-search condition to check.
        :param backtrack: a rule for calculating the next step-size to try.
        :param update_step_size: a rule for updating the step-size after each line-search.
        :param prox: a proximal operator. See 'proximal_ops'.
        :param mu: (optional) a lower-bound on the strong-convexity parameter of the objective.
            The method defaults to the parameter sequence for non-strongly convex functions when
            mu is not supplied.
        :param restart_rule: (optional) rule used to trigger a "restart" of the optimizer.
        """
        super().__init__(
            fista_ls,
            init_step_size,
            ls_cond,
            backtrack_fn,
            update_step_size,
            prox,
        )

        self.mu = mu
        self.v: Optional[lab.Tensor] = None
        self.t: float = 1.0
        self.restart_rule = restart_rule

        self.disp = None

    def reset(self):
        """Reset the secondary sequences to their initial state."""
        super().reset()
        # reset acceleration specific memory.
        self.restart()

    def restart(self):
        self.t = 1.0
        self.v = None

    def _check_restart(
        self, weights: lab.Tensor, old_weights: lab.Tensor, old_v: lab.Tensor
    ):
        if self.restart_rule is None or self.f0 is None or self.f1 is None:
            return
        # restart when function values increase
        elif self.restart_rule == F_VALS and self.f0 < self.f1:
            self.restart()
        # restart when step makes an obtuse angle with the gradient mapping.
        elif self.restart_rule == GM:
            if lab.sum((weights - old_weights) * (old_v - weights)) > 0.0:
                self.restart()

    def step(
        self,
        model: Model,
        X: lab.Tensor,
        y: lab.Tensor,
        f0: Optional[float] = None,
        grad: Optional[lab.Tensor] = None,
        batch_index: Optional[int] = None,
        batch_size: Optional[int] = None,
    ) -> Tuple[Model, Optional[float], Dict[str, Any]]:
        """Execute one step of an iterative optimization method.

        :param model: the model to update.
        :param X: the (n,d) data matrix to use in the update.
        :param y: the (n,) vector of targets.
        :param f0: (optional) current objective value. Can be provided to re-use computations.
        :param grad: (optional) current objective gradient. Can be provided to re-use computations.
        :param batch_index: NOT USED. The index of the current mini-batch.
        :param batch_size: the batch size to use when computing the objective.
            Defaults to `None' which indicates full-batch.
        :returns: the updated model and exit state of the line-search.
        """
        # compute objective and gradient without the non-smooth regularizer.
        obj_fn, grad_fn = model.get_closures(
            X, y, ignore_regularizer=True, batch_size=batch_size
        )

        if self.v is None:
            self.v = model.weights

        grad = grad_fn(self.v)
        descent_dir = grad

        # fv: function value at extrapolation.
        self.fv = obj_fn(self.v)
        # f0: function value at previous parameter
        self.f0 = self.f1
        old_weights = model.weights
        old_v = self.v

        # update model
        (
            model.weights,
            self.v,
            self.t,
            self.f1,
            self.step_size,
            exit_state,
        ) = self.step_fn(
            model.weights,
            self.fv,
            descent_dir,
            grad,
            obj_fn,
            grad_fn,
            self.step_size,
            self.ls_cond,
            self.backtrack_fn,
            self.prox,
            self.v,
            self.t,
            mu=self.mu,
        )

        # update step-size
        self.step_size = self.update_step_size(
            self.step_size,
            self.old_step_size,
            model.weights - old_v,
            self.fv,
            self.f1,
            grad,
        )
        self.old_step_size = self.step_size

        # check if we must restart the accelerated method.
        self._check_restart(model.weights, old_weights, old_v)

        return model, None, exit_state



"""Solvers for linear problems."""

import lab


def solve_ne(X: lab.Tensor, y: lab.Tensor, lam: float = 0.0) -> lab.Tensor:
    """Fit a linear model by directly solving the normal equations. This
    implementation forms X.T X and then solves the linear system.

    .. math::

        X^\top X w = X^\top b

    It is not memory efficient and should not be used when performance
    is required.
    :param X: (n,d) array containing the data examples.
    :param y: (n,d) array containing the data targets.
    :param lam: (optional) regularization parameter.
    :returns: the fit model.
    """
    n, d = X.shape

    # form X' X and X' y:
    XtX = X.T @ X
    Xty = X.T @ y

    return lab.solve(XtX + lam * lab.eye(d), Xty)



"""Pre-defined CVXPY programs.

TODO:
    - Look into merging some of the CVXPY code to reduce boilerplate.
"""

from .cvxpy_solver import CVXPYSolver

from .training_programs import (
    CVXPYGatedReLUSolver,
    CVXPYReLUSolver,
)

from .cone_decomposition import (
    MinL2Decomposition,
    MinL1Decomposition,
    FeasibleDecomposition,
    MinRelaxedL2Decomposition,
    SOCPDecomposition,
)


__all__ = [
    "CVXPYSolver",
    "CVXPYGatedReLUSolver",
    "CVXPYReLUSolver",
    "MinL2Decomposition",
    "MinL1Decomposition",
    "FeasibleDecomposition",
    "MinRelaxedL2Decomposition",
    "SOCPDecomposition",
]



"""Proximal gradient descent optimizer."""

from scnn.private.methods.core import (
    proximal_gradient_step,
    proximal_gradient_ls,
)
from scnn.private.methods.optimizers.proximal_optimizer import (
    ProximalOptimizer,
    ProximalLSOptimizer,
)
from scnn.private.methods.line_search import (
    StepSizeUpdater,
    LSCondition,
    Backtracker,
)
from scnn.private.prox import ProximalOperator


class PGD(ProximalOptimizer):

    """Proximal gradient descent (PGD) optimizer."""

    def __init__(
        self,
        step_size: float,
        prox: ProximalOperator,
        update_step_size: StepSizeUpdater,
    ):
        """
        :param step_size: the constant step-size to use.
        :param prox: a proximal operator. See 'proximal_ops'.
        :param update_step_size: a rule for updating the step-size after each line-search.
        """
        super().__init__(proximal_gradient_step, step_size, prox, update_step_size)


class PGDLS(ProximalLSOptimizer):
    """Proximal gradient descent with line-search."""

    def __init__(
        self,
        init_step_size: float,
        ls_cond: LSCondition,
        backtrack_fn: Backtracker,
        update_step_size: StepSizeUpdater,
        prox: ProximalOperator,
    ):
        """
        :param init_step_size: first step-size to try when running the line-search.
        :param ls_cond: the line-search condition to check.
        :param backtrack: a rule for calculating the next step-size to try.
        :param update_step_size: a rule for updating the step-size after each line-search.
        :param prox: a proximal operator. See 'proximal_ops'.
        """
        super().__init__(
            proximal_gradient_ls,
            init_step_size,
            ls_cond,
            backtrack_fn,
            update_step_size,
            prox,
        )



"""Solution mappings for the C-ReLU and C-GReLU problems."""
import torch

import lab

from scnn.private.models.convex import ConvexMLP
from scnn.private.models.non_convex import (
    GatedReLULayer,
    GatedReLUMLP,
    ReLUMLP,
)
from scnn.private.models.regularizers.l2 import L2Regularizer


def is_grelu_compatible(torch_model: torch.nn.Module) -> bool:
    submodules = list(torch_model.children())

    return (
        len(submodules) == 2
        and isinstance(submodules[0], GatedReLULayer)
        and submodules[0].linear.bias is None
        and isinstance(submodules[1], torch.nn.Linear)
        and submodules[1].bias is None
    )


def is_relu_compatible(torch_model: torch.nn.Module) -> bool:
    submodules = list(torch_model.children())

    return (
        len(submodules) == 3
        and isinstance(submodules[0], torch.nn.Linear)
        and submodules[0].bias is None
        and isinstance(submodules[1], torch.nn.ReLU)
        and isinstance(submodules[2], torch.nn.Linear)
        and submodules[2].bias is None
    )


def grelu_solution_mapping(convex_model, remove_sparse: bool = False):
    weights = convex_model.weights
    assert len(weights.shape) == 3

    weight_norms = (lab.sum(weights ** 2, axis=-1, keepdims=True)) ** (1 / 4)
    normalized_weights = lab.safe_divide(weights, weight_norms)

    first_layer = None
    second_layer = []
    for c in range(weights.shape[0]):
        pre_zeros = [
            lab.zeros_like(weight_norms[0]) for i in range(c)
        ]  # positive neurons
        post_zeros = [
            lab.zeros_like(weight_norms[0])
            for i in range(weights.shape[0] - c - 1)
        ]

        if first_layer is None:
            pre_weights = []
        else:
            pre_weights = [first_layer]

        first_layer = lab.concatenate(
            pre_weights
            + [
                normalized_weights[c],
            ],
            axis=0,
        )

        w2 = lab.concatenate(
            pre_zeros
            + [
                weight_norms[c],
            ]
            + post_zeros,
            axis=0,
        ).T
        second_layer.append(w2)
    second_layer = lab.concatenate(second_layer, axis=0)
    U = lab.concatenate(
        [convex_model.U for c in range(weights.shape[0])], axis=1
    )

    if remove_sparse:
        sparse_indices = lab.sum(first_layer, axis=-1) != 0

        first_layer = first_layer[sparse_indices]
        second_layer = second_layer[:, sparse_indices]

        U = U[:, sparse_indices]

    return (first_layer, second_layer, U)


def relu_solution_mapping(convex_model, remove_sparse: bool = False):
    weights = convex_model.weights
    assert len(weights.shape) == 4

    weight_norms = (lab.sum(weights ** 2, axis=-1, keepdims=True)) ** (1 / 4)
    normalized_weights = lab.safe_divide(weights, weight_norms)

    num_classes = weights.shape[1]
    first_layer = None
    second_layer = []
    for c in range(num_classes):
        pre_zeros = [
            lab.zeros_like(weight_norms[0, c]) for i in range(2 * c)
        ]  # positive neurons
        post_zeros = [
            lab.zeros_like(weight_norms[0, c])
            for i in range(2 * (num_classes - c - 1))
        ]

        if first_layer is None:
            pre_weights = []
        else:
            pre_weights = [first_layer]

        first_layer = lab.concatenate(
            pre_weights
            + [
                normalized_weights[0][c],
                normalized_weights[1][c],
            ],
            axis=0,
        )

        w2 = lab.concatenate(
            pre_zeros
            + [
                weight_norms[0][c],
                -weight_norms[1][c],
            ]
            + post_zeros,
            axis=0,
        ).T
        second_layer.append(w2)

    second_layer = lab.concatenate(second_layer, axis=0)

    if remove_sparse:
        sparse_indices = lab.sum(first_layer, axis=1) != 0

        first_layer = first_layer[sparse_indices]
        second_layer = second_layer[:, sparse_indices]

    return first_layer, second_layer


def convex_mlp_to_torch_mlp(
    convex_model: ConvexMLP,
    torch_model: torch.nn.Module,
    grelu: bool = False,
    remove_sparse: bool = False,
):
    submodules = list(torch_model.children())
    if grelu:
        assert is_grelu_compatible(torch_model)

        first_layer, second_layer, _ = grelu_solution_mapping(
            convex_model, remove_sparse
        )
        (submodules[0].linear.weight.data, submodules[1].weight.data) = (
            lab.torch_backend.torch_tensor(first_layer),
            lab.torch_backend.torch_tensor(second_layer),
        )
    else:
        assert is_relu_compatible(torch_model)

        first_layer, second_layer = relu_solution_mapping(
            convex_model, remove_sparse
        )

        (submodules[0].weight.data, submodules[2].weight.data,) = (
            lab.torch_backend.torch_tensor(first_layer),
            lab.torch_backend.torch_tensor(second_layer),
        )

    return torch_model


def construct_nc_torch(
    convex_model: ConvexMLP,
    grelu: bool = False,
    remove_sparse: bool = False,
):

    if grelu:
        first_layer, second_layer, U = grelu_solution_mapping(
            convex_model, remove_sparse
        )
        first_layer, second_layer, U = (
            lab.torch_backend.torch_tensor(first_layer),
            lab.torch_backend.torch_tensor(second_layer),
            lab.torch_backend.torch_tensor(U),
        )

        torch_model = torch.nn.Sequential(
            GatedReLULayer(U),
            torch.nn.Linear(U.shape[1], 1, bias=False),
        )
        submodules = list(torch_model.children())
        (submodules[0].linear.weight.data, submodules[1].weight.data,) = (
            first_layer,
            second_layer,
        )
    else:
        first_layer, second_layer = relu_solution_mapping(
            convex_model, remove_sparse
        )
        first_layer, second_layer = (
            lab.torch_backend.torch_tensor(first_layer),
            lab.torch_backend.torch_tensor(second_layer),
        )
        torch_model = torch.nn.Sequential(
            torch.nn.Linear(
                first_layer.shape[1], first_layer.shape[0], bias=False
            ),
            torch.nn.ReLU(),
            torch.nn.Linear(
                second_layer.shape[1], second_layer.shape[0], bias=False
            ),
        )
        submodules = list(torch_model.children())
        (submodules[0].weight.data, submodules[2].weight.data,) = (
            first_layer,
            second_layer,
        )

    return torch_model


def convex_mlp_to_manual_mlp(
    convex_model: ConvexMLP,
    manual_model: ReLUMLP,
    grelu: bool = False,
    remove_sparse: bool = False,
):
    if grelu:
        first_layer, second_layer, U = grelu_solution_mapping(
            convex_model, remove_sparse
        )
        assert isinstance(manual_model, GatedReLUMLP)

        manual_model.weights = manual_model._join_weights(
            first_layer, second_layer
        )
        manual_model.U = U
        manual_model.p = U.shape[1]
    else:
        first_layer, second_layer = relu_solution_mapping(
            convex_model, remove_sparse
        )
        manual_model.p = first_layer.shape[0]
        manual_model.weights = manual_model._join_weights(
            first_layer, second_layer
        )

    return manual_model


def construct_nc_manual(
    convex_model: ConvexMLP,
    grelu: bool = False,
    remove_sparse: bool = False,
):

    if grelu:
        manual_model = GatedReLUMLP(
            convex_model.d,
            convex_model.U,
            c=convex_model.c,
        )
    else:
        manual_model = ReLUMLP(
            convex_model.d, convex_model.p, c=convex_model.c
        )

    nc_model = convex_mlp_to_manual_mlp(
        convex_model, manual_model, grelu, remove_sparse
    )

    l2_regularizer = None
    if convex_model.regularizer is not None:
        l2_regularizer = L2Regularizer(convex_model.regularizer.lam)

    nc_model.regularizer = l2_regularizer

    return nc_model



"""Methods for solving linear systems."""

from .direct_solvers import solve_ne
from .iterative_solvers import lstsq_iterative_solve, linear_iterative_solve
from .preconditioners import get_preconditioner

__all__ = [
    "solve_ne",
    "lstsq_iterative_solve",
    "linear_iterative_solve",
    "get_preconditioner",
]



"""Gradient descent optimizer."""

from scnn.private.methods.core import gradient_step, gd_ls
from scnn.private.methods.optimizers.optimizer import Optimizer, LSOptimizer
from scnn.private.methods.line_search import StepSizeUpdater, LSCondition, Backtracker


class GD(Optimizer):

    """Simple (sub)gradient descent optimizer."""

    def __init__(
        self,
        step_size: float,
        update_step_size: StepSizeUpdater,
    ):
        """
        :param step_size: the constant step-size to use.
        :param update_step_size: a rule for updating the step-size after each line-search.
        """
        super().__init__(gradient_step, step_size, update_step_size)


class GDLS(LSOptimizer):

    """Gradient descent with line-search."""

    def __init__(
        self,
        init_step_size: float,
        ls_cond: LSCondition,
        backtrack_fn: Backtracker,
        update_step_size: StepSizeUpdater,
    ):
        """
        :param init_step_size: first step-size to try when running the line-search.
        :param ls_cond: the line-search condition to check.
        :param backtrack: a rule for calculating the next step-size to try.
        :param update_step_size: a rule for updating the step-size after each line-search.
        """
        super().__init__(gd_ls, init_step_size, ls_cond, backtrack_fn, update_step_size)



"""Approximate cone decompositions via one-sided quadratic.
"""
from typing import Optional, Tuple, List, Dict
from math import ceil

from scnn.private.models.model import Model
from scnn.private.models.regularizers import Regularizer

import lab


class QuadraticDecomposition(Model):

    """Quadratic approximation to the cone decomposition program."""

    def __init__(
        self,
        d: int,
        D: lab.Tensor,
        regularizer: Optional[Regularizer] = None,
        c: int = 1,
        combined: bool = True,
    ) -> None:
        """
        Args:
            d: the dimensionality of the dataset (ie. number of features).
            D: array of possible sign patterns.
            c: number of classes for this problem.
        """

        super().__init__(regularizer=regularizer)

        self.d = d
        self.D = D
        self.n, self.p = self.D.shape
        self.c = c

        self.weights = lab.zeros((self.c, self.p, self.d))

        self.orthant = 2 * self.D - lab.ones_like(self.D)

        self.combined = combined

    def update_model_weights(self, w: lab.Tensor):
        self.weights = w

    def _orthant(
        self, index_range: Optional[Tuple[int, int]] = None
    ) -> lab.Tensor:
        if index_range is None:
            return self.orthant

        return self.orthant[index_range[0] : index_range[1]]

    def _forward(
        self,
        X: lab.Tensor,
        w: lab.Tensor,
        orthant: Optional[lab.Tensor] = None,
        **kwargs,
    ) -> lab.Tensor:
        """Compute forward pass.

        Args:
            X: (n,d) array containing the data examples.
            w: parameter at which to compute the forward pass.
            orthant:
        """

        weights = self._weights(w).reshape(self.c, self.p, self.d)

        return lab.einsum("ij, mkj, ik-> mki", X, weights, orthant)

    def _objective(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: lab.Tensor,
        orthant: Optional[lab.Tensor] = None,
        scaling: Optional[float] = None,
        **kwargs,
    ) -> float:
        """Compute the l2 objective with respect to the model weights.

        Args:
            X: (n,d) array containing the data examples.
            y: (n,d) array containing the data targets.
            w: specific parameter at which to compute the objective.
            orthant: (optional) specific activation matrix at which to compute the forward pass.
                Defaults to self.D or manual computation depending on the value of self._train.
            scaling: (optional) scaling parameter for the objective. Defaults to `n * c`.
        """
        preds = self._forward(X, w, orthant)
        residual = lab.smax(y - preds, 0)
        obj = lab.sum(residual ** 2) / (2 * self._scaling(y, scaling))

        if not self.combined:
            obj += lab.sum(lab.smin(preds, 0) ** 2) / (
                2 * self._scaling(y, scaling)
            )

        return obj

    def _grad(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: lab.Tensor,
        orthant: Optional[lab.Tensor] = None,
        scaling: Optional[float] = None,
        flatten: bool = False,
        **kwargs,
    ) -> lab.Tensor:
        """Compute the gradient of augmented Lagrangian objective with respect
        to the model weights.

        Args:
            X: (n,d) array containing the data examples.
            y: (n,d) array containing the data targets.
            w: specific parameter at which to compute the objective.
            orthant: (optional) specific activation matrix at which to compute the forward pass.
                Defaults to self.D or manual computation depending on the value of self._train.
            scaling: (optional) scaling parameter for the objective. Defaults to `n * c`.
        """
        preds = self._forward(X, w, orthant)
        residual = lab.smax(y - preds, 0)

        grad = (
            -lab.einsum(
                "ij, mji, ik->mjk",
                orthant,
                residual,
                X,
            )
            / self._scaling(y, scaling)
        )

        if not self.combined:
            grad += (
                lab.einsum(
                    "ij, mji, ik->mjk",
                    orthant,
                    lab.smin(preds, 0),
                    X,
                )
                / self._scaling(y, scaling)
            )

        if flatten:
            return lab.ravel(grad)

        return grad

    def batch_X(
        self, batch_size: Optional[int], X: lab.Tensor
    ) -> List[Dict[str, lab.Tensor]]:

        if batch_size is None:
            return [{"X": X, "orthant": self.orthant}]
        n = X.shape[0]
        n_batches = ceil(n / batch_size)

        return [
            {
                "X": X[i * batch_size : (i + 1) * batch_size],
                "orthant": self.orthant[i * batch_size : (i + 1) * batch_size],
            }
            for i in range(n_batches)
        ]

    def batch_Xy(
        self, batch_size: Optional[int], X: lab.Tensor, y: lab.Tensor
    ) -> List[Dict[str, lab.Tensor]]:
        orthant = self.orthant

        if batch_size is None:
            return [{"X": X, "y": y, "orthant": self.orthant}]

        n = X.shape[0]
        n_batches = ceil(n / batch_size)

        return [
            {
                "X": X[i * batch_size : (i + 1) * batch_size],
                "y": y[i * batch_size : (i + 1) * batch_size],
                "orthant": orthant[i * batch_size : (i + 1) * batch_size],
            }
            for i in range(n_batches)
        ]



"""Iterative methods for solving linear systems and least-squares problems."""
from typing import Optional, Dict, Any, Tuple

import numpy as np
from scipy.sparse.linalg import LinearOperator, lsmr, lsqr, cg  # type: ignore


# methods

LSQR = "lsqr"
LSMR = "lsmr"
CG = "cg"

SOLVERS = [LSQR, LSMR, CG]

# parameters

TOL = 1e-6

# least-squares solvers.


def lstsq_iterative_solve(
    linear_op: LinearOperator,
    targets: np.ndarray,
    lam: float = 0.0,
    preconditioner: Optional[LinearOperator] = None,
    solver: str = LSMR,
    max_iters: int = 1000,
    tol: float = TOL,
) -> Tuple[np.ndarray, Dict[str, Any]]:
    """Use an iterative method to solve the l2-regularized least-squares
    problem,

    .. math::

        \|X w - y\|_2^2 + (\lambda/2) * \|w\|_2^2,

    where X is given by 'linear_op', y by 'targets', and lambda by 'lam'.
    WARNING: this function *only* supports NumPy and cannot be used with the generic backend
    in 'scnn.private.backend'.
    :param linear_op: a generalized linear-operator that evaluates matrix-vector products for
    the data matrix matrix X.
    :param targets: the response/targets y to predict.
    :param lam: (optional) the strength of the l2 regularization used. Defaults to no regularization.
    See [SciPy Documentation](https://docs.scipy.org/doc/scipy/reference/sparse.linalg.html) for details.
    :param preconditioner: (optional) a preconditioner to use when solving the system.
    See 'methods/linear/preconditioners.py'.
    :param solver: (optional) which solver to use. Valid options are LMSR and LSQR.
    :param max_iters: (optional) the maximum number of iterations to run the solver.
    :param tol: (optional) the tolerance to which the least-squares problem should be solved.
    :returns: (solution, exit_status) -- the solution to the least-squares problem and status of the solver.
    """
    assert isinstance(targets, np.ndarray)  # only support NumPy arrays.

    damping = 0.0
    if lam != 0.0:
        damping = np.sqrt(lam)

    if preconditioner is not None:
        # compose the operators.
        linear_op = linear_op.dot(preconditioner)

    if solver == LSMR:
        w_opt, istop, itn = lsmr(
            A=linear_op,
            b=targets,
            damp=damping,
            maxiter=max_iters,
            atol=tol,
            btol=tol,
        )[0:3]

        success = istop != 7
    elif solver == LSQR:
        w_opt, istop, itn, = lsqr(
            A=linear_op,
            b=targets,
            damp=damping,
            iter_lim=max_iters,
            atol=tol,
            btol=tol,
        )[0:3]

        success = istop != 7
    else:
        raise ValueError(f"Iterative solver {solver} not recognized!")

    # undo affect of preconditioning.
    if preconditioner is not None:
        w_opt = preconditioner.dot(w_opt)

    exit_status = {"success": success, "stop_condition": istop, "iterations": itn}

    return w_opt, exit_status


def linear_iterative_solve(
    linear_op: LinearOperator,
    targets: np.ndarray,
    preconditioner: Optional[LinearOperator] = None,
    solver: str = CG,
    max_iters: int = 1000,
    tol: float = TOL,
) -> Tuple[np.ndarray, Dict[str, Any]]:
    """Use an iterative method to solve the linear system.

    .. math::

        Xw = b,

    where X is given by 'linear_op', y by 'targets', and lambda by 'lam'.
    WARNING: this function *only* supports NumPy and cannot be used with the generic backend
    in 'scnn.private.backend'.
    :param linear_op: a generalized linear-operator that evaluates matrix-vector products for
    the data matrix matrix X.
    :param targets: the response/targets y to predict.
    :param preconditioner: (optional) a preconditioner to use when solving the system.
    See 'methods/linear/preconditioners.py'.
    :param solver: (optional) which solver to use. Valid options are CG.
    :param max_iters: (optional) the maximum number of iterations to run the solver.
    :returns: (solution, exit_status) -- the solution to the least-squares problem and status of the solver.
    """
    assert isinstance(targets, np.ndarray)  # only support NumPy arrays.

    if linear_op.shape[0] != linear_op.shape[1]:
        raise ValueError(
            "'linear_op' must specify a square matrix. Use 'lstsq_iterative_solve' for rectangular systems."
        )

    # flatten while solving the system.
    flat_targets = targets.reshape(-1)

    if solver == CG:
        w_opt, info = cg(
            A=linear_op,
            b=flat_targets,
            maxiter=max_iters,
            M=preconditioner,
            tol=tol,
            atol="legacy",
        )
        success = info == 0
    else:
        raise ValueError(f"Iterative solver {solver} not recognized!")

    exit_status = {"success": success}

    # restore original shape.
    w_opt = w_opt.reshape(targets.shape)

    return w_opt, exit_status



"""Utilities."""

from .linear_operators import MatVecOperator

from .linear import (
    solve_ne,
    lstsq_iterative_solve,
    linear_iterative_solve,
    get_preconditioner,
)

__all__ = [
    "solve_ne",
    "lstsq_iterative_solve",
    "linear_iterative_solve",
    "get_preconditioner",
]



"""Orthant constraint."""
from typing import Optional

import lab

from scnn.private.models.regularizers.constraint import Constraint

from scnn.private.prox import Orthant


class OrthantConstraint(Constraint):

    """Representation of the orthant constraint, A_i x >= 0, where A_i is a
    diagonal matrix with (A_i)_jk in {-1, 1}, as a regularizer.

    In essence, this is a wrapper for computation of the gradient mapping.
    """

    lam = 0.0

    def __init__(self, A: lab.Tensor):
        """
        :param A: a matrix of sign patterns defining orthants on which to project.
            The diagonal A_i is stored as the i'th column of A.
        :param lam: the tuning parameter controlling the strength of regularization.
        """

        self.projection_op = Orthant(A)



"""Non-convex models with manual implementations of the forward and backward
operators."""

from .relu_mlp import ReLUMLP
from .gated_relu_mlp import GatedReLUMLP

__all__ = [
    "ReLUMLP",
    "GatedReLUMLP",
]



"""Solvers for convex reformulations of neural networks based on the `CVXPY
<https://www.cvxpy.org>`_ DSL.

TODO:
    - How to handle GL1 penalties for the convex LassoNet models?
"""

from typing import Dict, Any, Tuple, List, Optional

from logging import root, INFO
import numpy as np
import cvxpy as cp

import lab

from scnn.private.models import (
    Model,
    Regularizer,
    L2Regularizer,
    GroupL1Regularizer,
    FeatureGroupL1Regularizer,
    ConvexMLP,
    AL_MLP,
)

from .cvxpy_solver import CVXPYSolver


class ConvexReformulationSolver(CVXPYSolver):
    """Solver for convex reformulations of neural network based on `CVXPY
    <https://www.cvxpy.org>`_.

    This is an abstract class that wraps CVXPY expressions used in all of the
    convex reformulations.
    """

    c: int
    d: int
    p: int
    n: int

    def process_inputs(
        self, model: ConvexMLP, X: np.ndarray, y: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Cast inputs into NumPy arrays and save problem dimensions."""

        # Cast variables to NumPy arrays.
        X_np = lab.to_np(X)
        y_np = lab.to_np(y)
        D_np = lab.to_np(model.D)

        # lookup problem dimensions
        self.n, self.d = X_np.shape
        self.c = y_np.shape[1]
        self.p = D_np.shape[1]

        return X_np, y_np, D_np

    def get_cone_constraints(
        self, W: cp.Variable, X: np.ndarray, A: np.ndarray
    ) -> List[cp.Expression]:
        """Form polyhedral cone constraints for the convex reformulation of a
        ReLU neural network.

        Args:
            W: a :math:`p * c \\times d` matrix of variables upon which to place the constraints,
                where `p` is the number of activation patterns, c is the target dimension, and `d`
                is the input dimension.
            X: a :math:`n \\times d` matrix of training examples.
            A: a :math:`d \\times p` matrix of signed activation patterns: :math:`A = 2 D - I`.

        Returns:
            A list of :math:`p * c` constraints, one for each of the target-activation pair.
        """
        return [
            cp.multiply(A, X @ W[i * self.p : (i + 1) * self.p].T) >= 0
            for i in range(self.c)
        ]

    def get_squared_error(
        self, W: cp.Variable, X: np.ndarray, y: np.ndarray, D: np.ndarray
    ) -> cp.Expression:
        """Form CVXPY expression for the squared-error in the convex
        reformulation.

        Args:
            W: a :math:`p * c \\times d` matrix of variables for the model weights,
                where `p` is the number of activation patterns, c is the target dimension, and `d`
                is the input dimension.
            X: a :math:`n \\times d` matrix of training examples.
            y: a :math:`n \\times c' matrix of targets.
            D: a :math:`d \\times p` matrix of activation patterns.

        Returns:
           An expression for the model's squared-error.
        """

        loss = 0.0
        for i in range(self.c):
            loss += (
                cp.sum_squares(
                    cp.sum(
                        cp.multiply(D, X @ W[i * self.p : (i + 1) * self.p].T),
                        axis=1,
                    )
                    - y[:, i]
                )
                / (2 * self.n * self.c)
            )

        return loss

    def get_regularization(
        self, W: cp.Variable, regularizer: Optional[Regularizer] = None
    ) -> cp.Expression:
        """Form CVXPY expression for the model regularization.

        Args:
            W: a :math:`p * c \\times d` matrix of variables for the model weights,
                where `p` is the number of activation patterns, c is the target dimension, and `d`
                is the input dimension.
            regularizer: the model regularizer.

        Returns:
           An expression for the regularization.
        """
        if regularizer is None or regularizer.lam == 0:
            return 0.0

        lam = regularizer.lam

        if isinstance(regularizer, L2Regularizer):
            return (lam / 2) * cp.sum_squares(W)
        elif isinstance(regularizer, GroupL1Regularizer):
            return lam * cp.mixed_norm(W, p=2, q=1)
        elif isinstance(regularizer, FeatureGroupL1Regularizer):
            return lam * cp.mixed_norm(W.T, p=2, q=1)
        else:
            raise ValueError(f"Regularizer {regularizer} not recognized!")


class CVXPYGatedReLUSolver(ConvexReformulationSolver):
    """CVXPY-based solver for convex Gated ReLU models."""

    def __call__(
        self, model: ConvexMLP, X: np.ndarray, y: np.ndarray
    ) -> Tuple[Model, Dict[str, Any]]:
        """Solve the convex reformulation for two-layer models with Gated ReLU
        activations.

        Args:
            model: the convex formulation to optimize.
            X: a :math:`n \\times d` matrix of training examples.
            y: a :math:`n \\times c' matrix of targets.

        Returns:
           The convex reformulation with optimized weights.
        """

        X_np, y_np, D_np = self.process_inputs(model, X, y)

        # create optimization variables
        W = cp.Variable((self.p * self.c, self.d))

        # get squared-error
        loss = self.get_squared_error(W, X_np, y_np, D_np)
        loss += self.get_regularization(W, model.regularizer)

        objective = cp.Minimize(loss)

        problem = cp.Problem(objective)

        # infer verbosity of sub-solver.
        verbose = root.level <= INFO
        # solve the optimization problem
        problem.solve(
            solver=self.solver, verbose=verbose, **self.solver_kwargs
        )

        # extract solution
        model.weights = lab.tensor(W.value, dtype=lab.get_dtype()).reshape(
            self.c, self.p, self.d
        )

        # extract solver information
        exit_status = {
            "success": problem.status == cp.OPTIMAL,
            "status": problem.status,
            "final_val": problem.value,
        }

        return model, exit_status


class CVXPYReLUSolver(ConvexReformulationSolver):
    """CVXPY-based solver for convex ReLU models."""

    def __call__(
        self, model: AL_MLP, X: np.ndarray, y: np.ndarray
    ) -> Tuple[Model, Dict[str, Any]]:
        """Solve the convex reformulation for two-layer models with ReLU
        activations.

        Args:
            model: the convex formulation to optimize.
            X: a :math:`n \\times d` matrix of training examples.
            y: a :math:`n \\times c' matrix of targets.

        Returns:
           The convex reformulation with optimized weights.
        """
        X_np, y_np, D_np = self.process_inputs(model, X, y)

        # create optimization variables
        U = cp.Variable((self.p * self.c, self.d))
        V = cp.Variable((self.p * self.c, self.d))
        W = U - V

        # get squared-error
        loss = self.get_squared_error(W, X_np, y_np, D_np)
        loss += self.get_regularization(U, model.regularizer)
        loss += self.get_regularization(V, model.regularizer)

        objective = cp.Minimize(loss)

        # define orthant constraints
        A = 2 * D_np - np.ones_like(D_np)
        constraints = self.get_cone_constraints(
            U, X_np, A
        ) + self.get_cone_constraints(V, X_np, A)

        problem = cp.Problem(objective, constraints)

        verbose = root.level <= INFO
        # solve the optimization problem
        problem.solve(
            solver=self.solver, verbose=verbose, **self.solver_kwargs
        )

        # extract solution
        model.weights = lab.stack(
            [
                lab.tensor(U.value, dtype=lab.get_dtype()).reshape(
                    self.c, self.p, self.d
                ),
                lab.tensor(V.value, dtype=lab.get_dtype()).reshape(
                    self.c, self.p, self.d
                ),
            ]
        )

        # extract solver information
        exit_status = {
            "success": problem.status == cp.OPTIMAL,
            "status": problem.status,
            "final_val": problem.value,
        }

        return model, exit_status



"""Construct efficient operators for convex neural networks."""
from typing import Tuple

from .kernels import einsum_kernel as ek
from .kernels import direct_kernel as dk

# constants

EINSUM = "einsum"
DIRECT = "direct"

KERNELS = [EINSUM, DIRECT]

# functions

expanded_data_matrix = dk.expanded_data_matrix
expanded_hessian = dk.expanded_hessian
expanded_bd_hessian = dk.expanded_bd_hessian


def get_kernel(kernel_name: str) -> Tuple:
    """Get a complete set of "kernel" functions for models.nns.ConvexMLP. These
    kernel functions implement operations with the expanded data matrix.

    .. math:: A = [D_1 X, D_2 X, ..., D_P X],
    without (necessarily) forming A directly. Note that these *only* work with the squared
    loss at the moment.
    :params kernel_name: the kernel to use. Current implemented kernels are
    EINSUM: compute operations without forming A using `einsum`; and
    DIRECT: form A and use the matrix directly.
    EINSUM is *much* more efficient than DIRECT and should be preferred except
    when running comparisons.
    :returns: tuple of Callable (data_mvp, data_t_mvp, grad, hessian_mvp, bd_hessian_mvp), where
    'data_mvp' computes products of the form :math:`A v`; and
    'data_t_mvp' computes transpose products A:math:`A.T w`; and
    'hessian_mvp' computes Hessian-vector products :math:`H v`; and
    'bd_hessian_mvp' computes Hessian-vector products with a block-diagonal
    approximation of the Hessian.
    """

    if kernel_name == EINSUM:
        kernel_funcs = einsum_kernel()
    elif kernel_name == DIRECT:
        kernel_funcs = direct_kernel()

    return kernel_funcs


def get_matrix_builders(kernel_name: str) -> Tuple:
    """Get a complete set of "matrix builders" for models.nns.ConvexMLP. These
    functions directly construct the matrices associated with the model, such
    as the expanded data matrix,

    .. math:: A = [D_1 X, D_2 X, ..., D_P X],
    without (necessarily) computing intermediate products. Note that these *only* work
    with the squared loss at the moment.
    :params kernel_name: the kernel to use. Current implemented kernels are
    EINSUM: compute matrices using `einsum`; and
    DIRECT: form the matrices by (slow) direct computation.
    EINSUM is *much* more efficient than DIRECT, which serves only as a reference
    implementation.
    :returns: tuple of Callable (data, hessian, bd_hessian), where
    'data_builder' computes the expanded data matrix :math:`A`;
    'hessian' computes the Hessian of the squared loss, :math:`A.T A`; and
    'bd_hessian' computes the diagonal blocks of the Hessian.
    """

    if kernel_name == EINSUM:
        builders = einsum_matrix_builders()
    elif kernel_name == DIRECT:
        builders = direct_matrix_builders()

    return builders


def einsum_kernel() -> Tuple:
    """Get "kernel" operators for a ConvexMLP with squared loss based on
    einsum.

    These operators are quite cryptic, but often an order of magnitude faster than
    direct computation.
    :returns: tuple of Callable (data_mvp, data_t_mvp, grad, hessian_mvp, bd_hessian_mvp), where
    'data_mvp' computes products of the form :math:`A v`;
    'data_t_mvp' computes transpose products :math:`A.T w`;
    'hessian_mvp' computes Hessian-vector products :math:`H v`; and
    'bd_hessian_mvp' computes Hessian-vector products with a block-diagonal
    approximation of the Hessian.
    """

    return (
        ek.data_mvp,
        ek.data_t_mvp,
        ek.gradient,
        ek.hessian_mvp,
        ek.bd_hessian_mvp,
    )


def einsum_matrix_builders() -> Tuple:
    """Get matrix builders for a ConvexMLP with squared loss based on einsum.

    These functions are quite cryptic, but often an order of magnitude faster than
    direct computation.
    :returns: tuple of Callable (data, hessian, bd_hessian), where
    'data_builder' computes the expanded data matrix :math:`A`;
    'hessian' computes the Hessian of the squared loss, :math:`A.T A`; and
    'bd_hessian' computes the diagonal blocks of the Hessian.
    """

    return ek.data, ek.hessian, ek.bd_hessian


# Direct implementations


def direct_matrix_builders():
    """Get matrix builders for a ConvexMLP with squared loss based on einsum.

    These functions are compute the matrices slowly and directly and are only
    meant as reference implementations.
    :returns: tuple of Callable (data, hessian, bd_hessian), where
    'data_builder' computes the expanded data matrix :math:`A`;
    'hessian' computes the Hessian of the squared loss, :math:`A.T A`; and
    'bd_hessian' computes the diagonal blocks of the Hessian.
    """

    return dk.data, dk.hessian, dk.bd_hessian


def direct_kernel():
    """Get "direct" operators for a ConvexMLP with squared loss based on
    einsum.

    These operators compute quantities by directly forming the expanded data matrix,
    which is slow but reliable. They are intended only for comparisons or as a reference.
    :returns: tuple of Callable (data_mvp, data_t_mvp, grad, hessian_mvp, bd_hessian_mvp), where
    'data_mvp' computes products of the form :math:`A v`;
    'data_t_mvp' computes transpose products :math:`A.T w`;
    'hessian_mvp' computes Hessian-vector products :math:`H v`; and
    'bd_hessian_mvp' computes Hessian-vector products with a block-diagonal
    approximation of the Hessian.
    """

    # pre-compute expanded matrix and hessian.

    return dk.data_mvp, dk.data_t_mvp, dk.gradient, dk.hessian_mvp, dk.bd_hessian_mvp



"""L2 Regularizer."""
from typing import Optional

import lab

from scnn.private.loss_functions import l1_squared_penalty
from scnn.private.models.regularizers.regularizer import Regularizer


class L1SquaredRegularizer(Regularizer):

    """L1-squared regularizer."""

    def penalty(self, w: lab.Tensor, **kwargs) -> float:
        """Compute the penalty associated with the regularizer.

        :param w: parameter at which to compute the penalty.
        :returns: penalty value
        """

        return l1_squared_penalty(w, self.lam)



"""Callback functions to be executed after each iteration of optimization."""
from typing import Tuple, Dict, Optional, Callable
import logging

import numpy as np

import lab

from scnn.private.models import (
    Model,
    Regularizer,
    QuadraticDecomposition,
    AL_MLP,
)
from scnn.private.methods.optimizers.pgd import PGDLS
from scnn.private.methods.optimization_procedures.iterative import (
    IterativeOptimizationProcedure,
)
from scnn.private.methods.optimizers import FISTA
from scnn.private.methods.termination_criteria import GradientNorm
from scnn.private.methods.line_search import (
    QuadraticBound,
    MultiplicativeBacktracker,
    Lassplore,
)
from scnn.private.prox import ProximalOperator


class ObservedSignPatterns:

    """Updates the set of active hyperplane arrangements by simultaneously
    running (S)GD on a ReLU MLP of fixed size."""

    def __init__(self):
        """
        :param relu_mlp: a ReLU MLP to be optimized concurrently with the convex model.
        :param optimizer: an optimizer for the ReLU MLP.
        """

        self.observed_patterns = {}
        self.hasher = hash

    def _get_hashes(self, D: lab.Tensor):
        return np.apply_along_axis(
            lambda z: self.hasher(np.array2string(z)), axis=0, arr=lab.to_np(D)
        )

    def _check_and_store_pattern(self, pattern) -> bool:
        if pattern in self.observed_patterns:
            return False
        else:
            self.observed_patterns[pattern] = True
            return True

    def __call__(self, model: Model, X: lab.Tensor, y: lab.Tensor) -> Model:

        # compute new sign patterns
        patterns, weights = model.sign_patterns(X, y)

        indices_to_keep = lab.tensor(
            [
                self._check_and_store_pattern(pattern)
                for pattern in self._get_hashes(patterns)
            ]
        )
        new_patterns = patterns[:, indices_to_keep]
        new_weights = weights[indices_to_keep]

        if model.activation_history is None or not hasattr(
            model, "activation_history"
        ):
            model.activation_history = new_patterns
            model.weight_history = new_weights
        else:
            model.activation_history = lab.concatenate(
                [model.activation_history, new_patterns], axis=1
            )
            model.weight_history = lab.concatenate(
                [model.weight_history, new_weights], axis=0
            )

        return model


class ConeDecomposition:

    """Convert a gated ReLU model into a ReLU model by solving the cone
    decomposition problem."""

    def __init__(self, solver: Callable):
        """
        :param solver: solver to use when computing the cone decomposition.
        """

        self.solver = solver

    def __call__(self, model: Model, X: lab.Tensor, y: lab.Tensor) -> Model:

        decomposed_model, _ = self.solver(model, X, y)

        return decomposed_model


class ApproximateConeDecomposition:

    """Convert a gated ReLU model into a ReLU model by approximating the cone
    decomposition problem."""

    def __init__(
        self,
        regularizer: Regularizer,
        prox: ProximalOperator,
        max_iters: int = 1000,
        tol: float = 1e-6,
        combined: bool = True,
    ):
        """
        Args:
            regularizer: the regularizer to use in the approximation.
            prox: the proximal operator to use for the given regularizer.
        """

        self.regularizer = regularizer
        self.prox = prox
        self.combined = combined

        # create optimizer to use
        self.optimizer = FISTA(
            1.0,
            QuadraticBound(),
            MultiplicativeBacktracker(beta=0.8),
            Lassplore(alpha=1.25, threshold=5.0),
            prox=self.prox,
        )

        # construct optimization routine
        self.opt_proc = IterativeOptimizationProcedure(
            self.optimizer,
            max_iters=max_iters,
            term_criterion=GradientNorm(tol),
            name="quadratic_decomp",
        )

    def __call__(self, model: Model, X: lab.Tensor, y: lab.Tensor) -> Model:

        # create decomposition to solve
        decomposition = QuadraticDecomposition(
            model.d,
            model.D,
            self.regularizer,
            c=model.c,
            combined=self.combined,
        )

        # form decomposition targets
        targets = -decomposition(X, model.weights)

        if self.combined:
            targets = lab.smax(targets, 0)

        # get logger object
        logger = logging.getLogger("scnn")

        # don't change initialization
        initializer = lambda x: x

        _, decomposition, _ = self.opt_proc(
            logger,
            decomposition,
            initializer,
            (X, targets),
            (X, targets),  # no test data, so pass train data.
            (["objective", "grad_norm"], [], []),
        )

        # compute decomposed weights
        w = decomposition.weights
        v = model.weights + w

        # extract ReLU network
        relu_model = AL_MLP(
            model.d,
            model.D,
            model.U,
            model.kernel,
            regularizer=model.regularizer,
            c=model.c,
        )
        relu_model.weights = lab.stack([v, w])

        return relu_model


class ProximalCleanup:
    """Cleanup the solution to an optimization problem by taking one proximal-
    gradient step."""

    def __init__(self, prox: ProximalOperator):
        """
        :param prox: the proximal-operator to use when cleaning-up the model solution.
        """
        self.prox = prox
        # create optimizer to use
        self.optimizer = PGDLS(
            1.0,
            QuadraticBound(),
            MultiplicativeBacktracker(beta=0.8),
            Lassplore(alpha=1.25, threshold=5.0),
            prox=self.prox,
        )

    def __call__(self, model: Model, X: lab.Tensor, y: lab.Tensor) -> Model:
        cleaned_model, _, exit_state = self.optimizer.step(model, X, y)

        return cleaned_model



"""Combined group L1 penalty and orthant constraints for the augmented
Lagrangian."""
from typing import Optional, Union, Callable

import lab

from scnn.private.models.regularizers.orthant import OrthantConstraint
from scnn.private.models.regularizers.group_l1 import GroupL1Regularizer

from scnn.private.prox import Orthant


class GroupL1Orthant(GroupL1Regularizer, OrthantConstraint):

    """Combined group L1 and orthant constraint for the augmented Lagrangian of
    a convex neural network."""

    def __init__(self, A: lab.Tensor, lam: float):
        """
        :param A: a matrix of sign patterns defining orthants on which to project.
            The diagonal A_i is stored as the i'th column of A.
        :param lam: the tuning parameter controlling the strength of regularization.
        """

        self.orthant_proj = Orthant(A)
        self.lam = lam

    def penalty(  # type: ignore
        self,
        w: lab.Tensor,
        split_weights: Callable[[lab.Tensor], lab.Tensor],
        **kwargs,
    ) -> float:
        """Compute the penalty associated with the regularizer.

        :param w: parameter at which to compute the penalty.
        :param split_weights: a function which can be called to split the weights
            into the group l1 terms and the orthant-constrained terms.
        :returns: penalty value
        """
        model_weights, _ = split_weights(w)

        return super().penalty(model_weights)

    def grad(
        self,
        w: lab.Tensor,
        base_grad: Optional[lab.Tensor] = None,
        split_weights: Callable[[lab.Tensor], lab.Tensor] = None,
        merge_weights: Callable[[lab.Tensor, lab.Tensor], lab.Tensor] = None,
        **kwargs,
    ) -> lab.Tensor:
        """Compute the gradient of the objective with respect to the model
        parameters.

        :param w: parameter at which to compute the penalty gradient.
        :param base_grad: (optional) the gradient of the un-regularized objective. This is
            used to compute the minimum-norm subgradient for "pseudo-gradient" methods.
        :param split_weights: a function which can be called to split the weights
            into the group l1 terms and the orthant-constrained terms.
        :param merge_weights: a function which can be called to merge the
            group-l1 germs and the orthant constrained terms into one parameter list.
        :returns: the gradient
        """
        assert base_grad is not None
        assert split_weights is not None
        assert merge_weights is not None

        model_weights, slacks = split_weights(w)
        model_weights_grad, slacks_grad = split_weights(base_grad)

        # compute min-norm subgradient for group-1 penalized variables.
        # super might not be reliable!
        min_norm_subgrad = model_weights_grad + super().grad(
            model_weights_grad, model_weights
        )

        # compute gradient mapping for orthant constrained variables.
        gradient_mapping = self._gradient_mapping(slacks, slacks_grad) - slacks_grad

        return merge_weights(min_norm_subgrad, gradient_mapping)



"""Test different methods for sampling activation patterns.

TODO:
    - 
"""

import unittest

import numpy as np

from scnn.private.utils.data.synthetic import gen_regression_data
from scnn.activations import (
    sample_gate_vectors,
    sample_sparse_gates,
    generate_index_lists,
    compute_activation_patterns,
)


class TestActivations(unittest.TestCase):
    """Test different methods for sampling activation patterns."""

    def setUp(self):
        # Generate realizable synthetic classification problem (ie. Figure 1)
        n_train = 100
        n_test = 100
        self.d = 25
        c = 2
        kappa = 1
        self.max_neurons = 100

        (
            (self.X_train, self.y_train),
            (self.X_test, self.y_test),
            _,
        ) = gen_regression_data(123, n_train, n_test, self.d, c, kappa=kappa)
        self.seed = 123
        self.n_gates = 1000

    def test_dense_gates(self):
        """Test sampling dense gate vectors."""

        G = sample_gate_vectors(
            self.seed, self.d, self.n_gates, gate_type="dense"
        )

        # check gate shape.
        self.assertTrue(G.shape == (self.d, self.n_gates))

        # check gates are dense
        self.assertTrue(np.all(G != 0.0))

    def test_feature_sparse_gates(self):
        """Test sampling feature-sparse gate vectors."""
        G = sample_sparse_gates(
            np.random.default_rng(self.seed),
            self.d,
            self.n_gates,
            sparsity_indices=np.arange(self.d).reshape(self.d, 1).tolist(),
        )

        # check gate shape.
        self.assertTrue(G.shape == (self.d, self.n_gates))

        # each gate should have exactly one non-zero index
        self.assertTrue(np.all(np.sum(G != 0, axis=0) == 1))

        # try generating index lists
        order_one = generate_index_lists(self.d, 1)
        self.assertTrue(
            np.all(order_one == np.arange(self.d).reshape(self.d, 1).tolist())
        )

        order_two = generate_index_lists(self.d, 2)

        G = sample_sparse_gates(
            np.random.default_rng(self.seed),
            self.d,
            self.n_gates,
            sparsity_indices=order_two,
        )

        # each gate should have exactly one or two non-zero indices
        n_sparse_indices = np.sum(G != 0, axis=0)
        self.assertTrue(
            np.all(np.logical_or(n_sparse_indices == 1, n_sparse_indices == 2))
        )

    def test_bias_terms(self):
        """Test sampling gate vectors with bias terms."""

        # manually compute activations:
        def activations(X, G):
            XG = np.matmul(X, G)
            XG = np.maximum(XG, 0)
            XG[XG > 0] = 1
            return XG

        # augment data with bias term
        X = np.concatenate(
            [self.X_train, np.ones((self.X_train.shape[0], 1))], axis=1
        )

        G = sample_gate_vectors(
            123,
            self.d,
            self.n_gates,
            gate_type="dense",
        )

        D, G_bias = compute_activation_patterns(
            X, G, bias=True, active_proportion=0.5
        )

        self.assertTrue(
            np.all(np.sum(D, axis=0) == X.shape[0] / 2),
            "Half of the examples should be active for each hyperplane",
        )

        D_manual = activations(X, G_bias)
        self.assertTrue(
            np.all(D_manual == D), "The activations should be produced by G."
        )

        D, G_bias = compute_activation_patterns(
            X, G, bias=True, active_proportion=0.1
        )

        self.assertTrue(
            np.all(np.sum(D, axis=0) == X.shape[0] / 10),
            "1/10 of the examples should be active for each hyperplane",
        )

        D_manual = activations(X, G_bias)
        self.assertTrue(
            np.all(D_manual == D), "The activations should be produced by G."
        )


if __name__ == "__main__":
    unittest.main()



"""Termination criteria for optimization methods.

TODO:
    - Update `StepLength` to be the gradient mapping. The major
        difference is multiplication by a step-size.
"""

from typing import Optional

import lab
from scnn.private.models.model import Model


class TerminationCriterion:
    """Base class for termination criteria.

    A boolean function with one or more tuning parameters that returns `True`
    when an optimization procedure should terminate and `False` otherwise.

    Attributes:
        tol: a parameter controlling sensitivity of the termination criterion.
    """

    tol: float

    def __init__(self, tol: float):
        """
        Args:
            tol: a tolerance parameter controlling sensitivity of the
                termination criterion.
        """

        self.tol = tol

    def __call__(
        self,
        model: Model,
        X: lab.Tensor,
        y: lab.Tensor,
        objective: Optional[lab.Tensor] = None,
        grad: Optional[lab.Tensor] = None,
    ) -> bool:
        """Evaluate the termination criterion given a model and a dataset.

        The current objective value and gradient are optional parameters;
        these should be supplied if they have been pre-calculated for another
        purpose. Otherwise, the criterion will compute them as necessary.

        Args:
            model: the prediction model that is being optimized.
            X: a :math:`n \\times d` matrix of training examples.
            y: a :math:`n \\times c` matrix of training targets.
            objective: the current objective value.
                Provide only if already computed.
            grad: the current gradient of the objective.
                Provide only if already computed.

        Returns:
            Boolean indicating whether or not optimization should terminate.
        """
        raise NotImplementedError(
            "Termination criteria must implement __call__!"
        )


class ConstrainedTerminationCriterion(TerminationCriterion):
    """Convergence criterion for constrained optimization problems.

    Base class for termination criteria for problems with constraints.
    Provides functions for computing the constraint violations.

    Attributes:
        obj_tol: the tolerance for the primal objective.
        constraint_tol: the tolerance for violation of the constraints.
    """

    obj_tol: float
    constraint_tol: float

    def __init__(self, obj_tol: float, constraint_tol: float):
        """
        Args:
            obj_tol: the tolerance for determining optimality of the primal
                optimization problem.
            constraint_tol: the tolerance for determining feasibility.
        """
        self.obj_tol = obj_tol
        self.constraint_tol = constraint_tol

    def constraint_violations(self, model: Model, X: lab.Tensor) -> float:
        """Compute current constraint violations.

        Args:
            model: the prediction model that is being optimized.
            X: a :math:`n \\times d` matrix of training examples.

        Returns:
            Squared l2-norm of constraint violations.
        """

        e_gap, i_gap = model.constraint_gaps(X)
        return lab.sum(e_gap ** 2 + lab.smax(i_gap, 0) ** 2)


class GradientNorm(TerminationCriterion):
    """First-order optimality criterion.

    Terminate optimization if and only if the norm of minimum-norm subgradient
    is below a certain tolerance.


    Attributes:
        tol: the tolerance for the gradient norm. The objective is
            approximately stationary if the gradient norm is less than `tol`.
    """

    def __call__(
        self,
        model: Model,
        X: lab.Tensor,
        y: lab.Tensor,
        objective: Optional[lab.Tensor] = None,
        grad: Optional[lab.Tensor] = None,
    ) -> bool:
        """Terminate if gradient norm is sufficiently small.

        Determine if the norm of the minimum-norm sub-gradient (or gradient if
        the function is smooth) is small enough (according to self.tol) to
        constitute a first-order stationary point.

        The current objective value and gradient are optional parameters;
        these should be supplied if they have been pre-calculated for another
        purpose. Otherwise, the criterion will compute them as necessary.

        Args:
            model: the prediction model that is being optimized.
            X: a :math:`n \\times d` matrix of training examples.
            y: a :math:`n \\times c` matrix of training targets.
            objective: the current objective value. NOT USED.
            grad: the current gradient of the objective.
                Provide only if already computed.

        Returns:
            Boolean indicating whether or not optimization should terminate.
        """

        if grad is None:
            grad = model.grad(X, y)

        return lab.sum(grad ** 2) <= self.tol


class StepLength(TerminationCriterion):
    """Criterion based on length of the most recent step.

    Terminate optimization if and only if the norm of the last step was
    below a certain tolerance.

    Notes:
        This criterion can significant increase the memory requirements
        of an optimization procedure due to the `previous_weights` attributes.

    Attributes:
        previous_weights: the parameters of the model from the previous
            iteration. These are necessary to compute the length of the
            step.

    """

    previous_weights: Optional[lab.Tensor] = None

    def __call__(
        self,
        model: Model,
        X: lab.Tensor,
        y: lab.Tensor,
        objective: Optional[lab.Tensor] = None,
        grad: Optional[lab.Tensor] = None,
    ) -> bool:
        """Terminate if step-length is sufficiently small.

        Determine if the norm of previous step is small enough (according to
        `self.tol`) to indicate the method has converged.

        The current objective value and gradient are not used.

        Args:
            model: the prediction model that is being optimized.
            X: a :math:`n \\times d` matrix of training examples.
            y: a :math:`n \\times c` matrix of training targets.
            objective: the current objective value. NOT USED.
            grad: the current gradient of the objective. NOT USED.

        Returns:
            Boolean indicating whether or not optimization should terminate.
        """
        if self.previous_weights is None:
            self.previous_weights = model.weights
            return False

        step = model.weights - self.previous_weights
        self.previous_weights = model.weights

        return lab.sqrt(lab.sum(step ** 2)) <= self.tol


class ConstrainedHeuristic(ConstrainedTerminationCriterion):
    """Terminate if the gradient norm and constraint violations are both small.

    A heuristic condition which terminates optimization if and only if the
    norm of minimum-norm subgradient is below a certain tolerance and the norm
    of the constraints violations is below a separate tolerance.

    Notes:
        - The gradient norm must be computed with respect a penalized
            objective; otherwise, termination is only possible if a solution
            exists in the interior of the constraint set.

        - This criterion is not the same as checking stationarity of the
            Lagrangian, but appears to work well in practice for penalized
            objectives.

    Attributes:
        obj_tol: the tolerance for the gradient norm. The penalized objective
            is approximately stationary if the gradient norm is less than
            `grad_tol`.
        constraint_tol: the tolerance for violation of the constraints.
            The model is approximately feasible if the norm of the constraint
            violations is less than `constraint_tol`.
    """

    obj_tol: float
    constraint_tol: float

    def __call__(
        self,
        model: Model,
        X: lab.Tensor,
        y: lab.Tensor,
        objective: Optional[lab.Tensor] = None,
        grad: Optional[lab.Tensor] = None,
    ) -> bool:
        """Terminate if gradient norm and constraint violations are small.

        Args:
            model: the prediction model that is being optimized.
            X: a :math:`n \\times d` matrix of training examples.
            y: a :math:`n \\times c` matrix of training targets.
            objective: the current objective value. NOT USED.
            grad: the current gradient of the objective.
                Provide only if already computed.

        Returns:
            Boolean indicating whether or not optimization should terminate.
        """
        if grad is None:
            grad = model.grad(X, y)

        gaps = self.constraint_violations(model, X)

        if gaps <= self.constraint_tol:
            return lab.sum(grad ** 2) <= self.obj_tol

        return False


class LagrangianGradNorm(ConstrainedTerminationCriterion):
    """First-order optimality criterion for primal-dual methods.

    Terminate optimization if and only if the norm of minimum-norm subgradient
    of the Lagrangian function is below a certain tolerance and the current
    point is approximately feasible. This criterion is only supported for
    solvers which maintain both primal and dual parameters. The gradient is
    computed with respect the primal parameters.

    Attributes:
        obj_tol: the tolerance for the gradient norm. The Lagrangian
            is approximately stationary if the gradient norm is less than
            `grad_tol`.
        constraint_tol: the tolerance for violation of the constraints.
            The model is approximately feasible if the norm of the constraint
            violations is less than `constraint_tol`.
    """

    obj_tol: float
    constraint_tol: float

    def __call__(
        self,
        model: Model,
        X: lab.Tensor,
        y: lab.Tensor,
        objective: Optional[lab.Tensor] = None,
        grad: Optional[lab.Tensor] = None,
    ) -> bool:
        """Terminate if the Lagrangian is approximately stationary and the
        constraint violations are small.

        Args:
            model: the prediction model that is being optimized.
            X: a :math:`n \\times d` matrix of training examples.
            y: a :math:`n \\times c` matrix of training targets.
            objective: the current objective value. NOT USED.
            grad: the current gradient of the objective. NOT USED.

        Returns:
            Boolean indicating whether or not optimization should terminate.
        """

        gaps = self.constraint_violations(model, X)
        if gaps <= self.constraint_tol:
            return lab.sum(model.lagrangian_grad(X, y) ** 2) <= self.obj_tol

        return False
